import inspect
import uuid
from dataclasses import dataclass
from itertools import chain
from pprint import pformat
from typing import Callable, List, Any, Dict, Union

import networkx as nx
from cytoolz import memoize
from loguru import logger
from pampy import match
from pinject.bindings import default_get_arg_names_from_class_name
from pinject.finding import find_classes
from pyvis.network import Network
from returns.result import safe, Result, Failure

from pinject_design.di.design import PinjectBind, Bind, InjectedProvider
from pinject_design.di.injected import Injected, InjectedFunction, InjectedPure, MappedInjected, \
    ZippedInjected, MZippedInjected, InjectedByName, extract_dependency
from pinject_design.di.util import Design, DirectPinjectProvider, PinjectProviderBind
from pinject_design.exceptions import DependencyResolutionFailure, _MissingDepsError


def dfs(neighbors: Callable, node: str, trace=[]):
    nexts: List[str] = neighbors(node)
    for n in nexts:
        yield node, n, trace
        yield from dfs(neighbors, n, trace + [n])


safe_attr = safe(getattr)


@safe
def getitem(tgt, name):
    return tgt[name]


# %%
@dataclass
class PinProvider:
    src: Callable

    def __post_init__(self):
        # self.arg_binding_keys:Result = safe_attr(self.src, "_pinject_arg_binding_keys")
        self.non_injectables: Result = safe_attr(self.src, "_pinject_non_injectables")
        # self.orig_f:Result = safe_attr(self.src, "_pinject_orig_fn")
        # self.provider_decorations:Result = safe_attr(self.src, "_pinject_provider_decorations")


@dataclass
class DIGraph:
    src: Design

    def _get_configured(self):
        for k, b in self.src.bindings.items():
            if isinstance(b, DirectPinjectProvider) and hasattr(b.method, "_pinject_is_wrapper"):
                pp = PinProvider(b.method)
                yield k, pp

    def _get_mapping(self):
        classes = find_classes(self.src.modules, self.src.classes)
        for c in classes:
            for name in default_get_arg_names_from_class_name(c.__name__):
                yield name, c

    def new_name(self, base: str):
        return f"{base}_{str(uuid.uuid4())[:6]}"

    def __post_init__(self):
        self.src = self.src.build()
        self.implicit_mappings = dict(self._get_mapping())
        self.pinject_mappings = dict(self._get_configured())
        # we want to know if the binding is InjectedProvider or not
        self.explicit_mappings: Dict[str, Bind] = {k: b for k, b in self.src.bindings.items() if
                                                   k not in self.pinject_mappings}
        self.multi_mappings = {k: b for k, b in self.src.multi_binds.items()}

        self.direct_injected = dict()
        self.injected_to_id = dict()

        @memoize
        def deps_impl(src: str):
            if "provide_" in src:
                src = src.replace("provide_", "")
            if src in self.explicit_mappings:
                em = self.explicit_mappings[src]
                if isinstance(em, InjectedProvider):
                    return self.resolve_injected(em.src)
                else:
                    return em.to_injected().dependencies()
            elif src in self.implicit_mappings:
                return Injected.bind(self.implicit_mappings[src]).dependencies()
            elif src in self.pinject_mappings:
                pp: PinProvider = self.pinject_mappings[src]
                deps = [d for d in Injected.bind(pp.src).dependencies() if
                        d not in pp.non_injectables.value_or([])]
                return deps
            elif src in self.multi_mappings:
                return list(set(chain(*[Injected.bind(tgt).dependencies() for tgt in self.multi_mappings[src]])))
            elif src in self.direct_injected:
                di = self.direct_injected[src]
                return self.resolve_injected(di)
            else:
                raise RuntimeError(f"key not found!:{src}")

        self.deps_impl = deps_impl

    def resolve_injected(self, i: Injected) -> List[str]:
        "give new name to unknown manual injected valus and return dependencies"
        if i not in self.injected_to_id:
            self.injected_to_id[i] = str(uuid.uuid4())[:6]
            # logger.info(f"add injected node:\n{tabulate.tabulate(list(self.injected_to_id.items()))}")
        #        else:
        #            assert False
        _id = self.injected_to_id[i]
        if isinstance(i, InjectedFunction):
            deps = []
            for k, v in i.kwargs_mapping.items():
                dep_name = f"{k}_{_id}"
                injected = match(v,
                                 str, Injected.by_name,
                                 type, Injected.bind,
                                 callable, Injected.bind,
                                 Injected, lambda i: i
                                 )
                self.direct_injected[dep_name] = injected

                # logger.info(f"add direct injected node:\n{tabulate.tabulate(list(self.direct_injected.items()))}")
                deps.append(dep_name)
            missings = list(sum([list(extract_dependency(m)) for m in i.missings], []))
            return deps + list(missings)
        elif isinstance(i, MappedInjected):
            dep_name = self.new_name("mapped_src")
            # dep_name = f"mapped_injected_src_{str(uuid.uuid4())[:6]}"
            self.direct_injected[dep_name] = i.src
            return [dep_name]
        elif isinstance(i, ZippedInjected):
            a = self.new_name("zipped_src_a")
            b = self.new_name("zipped_src_b")
            self.direct_injected[a] = i.a
            self.direct_injected[b] = i.b
            return [a, b]
        elif isinstance(i, MZippedInjected):
            res = []
            for k, src in enumerate(i.srcs):
                sn = self.new_name(f"mzip_src_{k}")
                self.direct_injected[sn] = src
                res.append(sn)
            return res
        else:
            return list(i.dependencies())

    def __getitem__(self, key):
        if "provide_" in key:
            key = key.replace("provide_", "")
        item = getitem(
            self.src.bindings, key
        ).lash(
            lambda e: getitem(self.implicit_mappings, key)
        ).lash(
            lambda e: getitem(self.pinject_mappings, key)
        ).lash(
            lambda e: getitem(self.multi_mappings, key)
        ).lash(
            lambda e: getitem(self.direct_injected, key)
        )
        if isinstance(item, Failure):
            raise KeyError(f"{key} is not found in design!")
        return item.unwrap()

    def dependencies_of(self, src):
        return self.deps_impl(src)

    def di_dfs(self, src):
        def dfs(node: str, trace=[]):
            try:
                nexts: List[str] = self.dependencies_of(node)
            except Exception as e:
                raise _MissingDepsError(f"failed to get neighbors of {node} at {' => '.join(trace)}.", node,
                                        trace) from e
            for n in nexts:
                yield node, n, trace
                yield from dfs(n, trace + [n])

        yield from dfs(src, [src])

    def di_dfs_validation(self, src):
        def dfs(node: str, trace=[]):
            nexts = []
            try:
                nexts: List[str] = self.dependencies_of(node)
            except Exception as e:
                yield DependencyResolutionFailure(node, trace, e)
            for n in nexts:
                yield from dfs(n, trace + [n])

        yield from dfs(src, [src])

    def create_dependency_network(self, roots: Union[str, List[str]]):
        if isinstance(roots, str):
            roots = [roots]

        def get_source(f):
            try:
                file = inspect.getfile(f)
                src = inspect.getsource(f)
                res = file + "\n" + src

            except Exception as e:
                logger.warning(f"{e} error from {f}")
                res = f"failed:{f} from {f}"

            res = res.replace("<", "").replace(">", "")
            return res

        # get_source = lambda f: safe(inspect.getsource)(f).value_or(f"failed:{f}").replace("<", "").replace(">", "")

        nx_graph = nx.DiGraph()
        for root in roots:
            for a, b, trc in self.di_dfs(root):
                nx_graph.add_edge(b, a)

        @memoize
        def node_to_sl(n):
            def parse(tgt):
                return match(tgt,
                             Injected, parse_injected,
                             InjectedProvider(Injected), parse_injected,
                             PinjectProviderBind, lambda ppb: ("function", ppb.f.__name__, get_source(ppb.f)),
                             DirectPinjectProvider, lambda dpp: ("method", dpp.method.__name__, get_source(dpp.method)),
                             PinjectBind({"to_instance": callable}), lambda i: ("instance", str(i), get_source(i)),
                             PinjectBind({"to_instance": Any}),
                             lambda i: ("instance", str(i), f"instance:{pformat(i)}"),
                             PinjectBind({"to_class": Any}), lambda i: ("class", i.__name__, get_source(i)),
                             type, lambda cls: ("class", cls.__name__, get_source(cls)),
                             list, lambda providers: ("multi_binding", repr(providers), repr(providers)),
                             # MetaBind, lambda mb: (mb.metadata["src"],parse(mb.src)[1],mb.metadata["src"]),
                             Any, lambda a: ("unknown", str(a), f"unknown-type:{type(a)}=>{a}")
                             )

            def parse_injected(tgt: Injected):
                # from archpainter.my_artifacts.artifact_object import IArtifactObject
                return match(tgt,
                             InjectedFunction,
                             lambda injected: ("injected",
                                               f"Injected:{safe(getattr)(injected.target_function, '__name__').value_or(repr(injected.target_function))}",
                                               get_source(injected.target_function)),
                             InjectedPure,
                             lambda injected: (
                                 "injected",
                                 f"Pure:{injected.value}",
                                 get_source(injected.value)
                                 if isinstance(injected, Callable)
                                 else str(injected.value)
                             ),
                             MappedInjected,
                             lambda injected: ("injected", f"{injected.__class__.__name__}", get_source(injected.f)),
                             ZippedInjected,
                             lambda injected: ("injected", f"{injected.__class__.__name__}", "zipped"),
                             MZippedInjected,
                             lambda injected: ("injected", f"{injected.__class__.__name__}", "mzipped"),
                             # IArtifactObject,
                             # lambda injected: ("injected", f"artifact:{injected.metadata.identifier}", str(injected)),
                             InjectedByName,
                             lambda injected: (
                                 "injected", f"name:{injected.name}", f"injected by name:{injected.name}"),
                             Injected,
                             lambda injected: ("injected", f"{injected.__class__.__name__}", str(injected)),
                             )

            group, short, long = parse(self[n])
            short = str(short).replace("<", "").replace(">", "")[:100]
            # long = long.replace("\n", "<br>")  # .replace("<","").replace(">","")

            # logger.debug(short)
            # logger.debug(long)
            n_edges = len(list(nx_graph.neighbors(n))) + len(self.dependencies_of(n))
            return dict(
                label=f"{n}\n{short}",
                title=long,
                group=group,
                value=n_edges,
                mass=n_edges * 0.5 + 1,
            )

        for root in roots:
            nx_graph.add_node(root)
        for n in nx_graph.nodes:
            assert isinstance(n, str)
            attrs = node_to_sl(n)
            node = nx_graph.nodes[n]
            for k, v in attrs.items():
                node[k] = v
        for root in roots:
            nx_graph.nodes[root]["group"] = "root"

        nt = Network('100%', '100%', directed=True)
        nt.from_nx(nx_graph)
        nt.show_buttons(filter_=["physics"])
        nt.toggle_physics(True)
        return nt

    def find_missing_dependencies(self, roots: Union[str, List[str]]) -> List[DependencyResolutionFailure]:
        if isinstance(roots, str):
            roots = [roots]
        failures = []
        for r in roots:
            failures += self.di_dfs_validation(r)
        return failures

    def show_graph_notebook(self, roots: Union[str, List[str]]):
        nt = self.create_dependency_network(roots)
        nt.width = 1000
        nt.height = 1000
        nt.prep_notebook()
        return nt.show("__notebook__.html")

    def save_as_html(self, roots: Union[str, List[str]], dst_path: str):
        nt = self.create_dependency_network(roots)
        nt.show(dst_path)


# %%
def create_dependency_graph(d: Design, roots: List[str], output_file="dependencies.html"):
    dig = DIGraph(d.bind_instance(
        job_type="net_visualization"
    ))
    nt = dig.create_dependency_network(roots)
    nt.show(output_file)
    return nt

# %%
# import os
#
# d = find_cfg_by_alias("ceylon").exp_design()
# create_dependency_graph(d, "rgb_to_xyz_training", "out.html")
#
# os.system("open out.html")



import abc
import inspect
from dataclasses import dataclass
from functools import wraps
from typing import TypeVar, Generic, Callable, Union, Dict, List

from frozendict import frozendict
from loguru import logger
from makefun import create_function

from pinject_design.di.injected import Injected, extract_dependency_including_self, extract_dependency

T = TypeVar("T")
U = TypeVar("U")


@dataclass
class PinjectConfigure:
    kwargs: Dict  # a dict like:{to_class=A} / {to_instance=0}

    def __hash__(self):
        return hash(frozendict(**self.kwargs))


@dataclass
class PinjectProvider:
    method: Callable  # a callable with (self,dep1,dep2) signatures...

    # might be annotated with pinject's annotations..
    # pinject ads an attribute to an annotated function for marking..
    # which is not something I prefer though...
    # ok I should stop using pinject's features

    def __post_init__(self):
        # if self is not in method signature, this will add it.
        assert self.method is not None, "PinjectProvider created with method==None!"
        if callable(self.method):
            self.method = ensure_self_arg(self.method)
            # we need to make sure the function's name starts with provide_
        else:
            # may this class is used for pampy pattern matching..
            pass

    def __hash__(self):
        return hash(self.method)


class Bind(Generic[T], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def to_pinject_binding(self) -> Union[PinjectConfigure, PinjectProvider]:
        pass

    def map(self, f: Callable[[T], U]) -> "Bind[U]":
        return MappedBind(self, f)

    @staticmethod
    def provider(f: Callable):
        return FunctionProvider(f)

    @staticmethod
    def injected(injected: Injected):
        return InjectedProvider(injected)

    @staticmethod
    def instance(instance: object):
        return PinjectBind(kwargs=dict(to_instance=instance))

    @staticmethod
    def clazz(cls: type):
        return PinjectBind(kwargs=dict(to_class=cls))

    def to_injected(self) -> Injected:
        return bind_to_injected(self)


def map_wrap(src, f):
    @wraps(src)
    def wrapper(*args, **kwargs):
        return f(src(*args, **kwargs))

    wrapper.__signature__ = inspect.signature(src)
    return wrapper


def pinject_to_provider(binding: Union[PinjectConfigure, PinjectProvider]):
    from pampy import match, _
    new_provider = match(binding,
                         PinjectConfigure({"to_class": _}), lambda cls: cls,  # create a function provider
                         PinjectConfigure({"to_instance": _}), lambda item: lambda: item,
                         PinjectProvider(_), lambda f: f
                         )
    return new_provider


@dataclass
class MappedBind(Bind[U]):
    src: Bind[T]
    f: Callable[[T], U]

    def to_pinject_binding(self) -> Union["PinjectConfigure", "PinjectProvider"]:
        binding = self.src.to_pinject_binding()
        new_provider = pinject_to_provider(binding)
        mapped = map_wrap(new_provider, self.f)
        return PinjectProvider(mapped)


def remove_kwargs_from_func(f, kwargs: List[str]):
    deps = extract_dependency_including_self(f)
    to_remove = set(kwargs)
    new_kwargs = deps - to_remove
    func_name = f.__name__.replace("<lambda>", "_lambda_")
    sig = f"""{func_name}({",".join(new_kwargs)})"""

    def impl(**called_kwargs):
        deleted = deps & to_remove
        # for self, you must check whether f is method or not
        if inspect.ismethod(f):
            deleted = deleted - {"self"}
        d_kwargs = {k: None for k in deleted}
        return f(**called_kwargs, **d_kwargs)

    return create_function(sig, impl)


def bind_to_injected(bind: Bind):
    if isinstance(bind, PinjectBind) and "to_class" in bind.kwargs:
        cls = bind.kwargs["to_class"]
        return Injected.bind(cls)
    elif isinstance(bind, InjectedProvider):
        return bind.src
    pb = bind.to_pinject_binding()
    provider = pinject_to_provider(pb)
    provider = remove_kwargs_from_func(provider, ["self"])
    return Injected.bind(provider)


# jetbrains://idea/navigate/reference?project=archpainter&path=notes/organized/mac/check_fid_and_generated.ipynb

# A contract of Bind is that it is possible to return
# Configure/Provider to be assigned to BindingSpec given a key name.


# in order to make this class picklable, we need to postpone the wrapping process until to_binding_spec() call.
class ProviderTrait(Bind[T]):
    @property
    @abc.abstractmethod
    def provider(self):
        pass

    def to_pinject_binding(self) -> Union["PinjectConfigure", "PinjectProvider"]:
        provider = self.provider
        assert provider is not None, "provider is None for some reason!"
        return PinjectProvider(provider)


@dataclass
class InjectedProvider(ProviderTrait[T]):
    src: Injected[T]

    @property
    def provider(self):
        return self.src.get_provider()


@dataclass
class FunctionProvider(ProviderTrait[T]):
    f: Callable

    def __post_init__(self):
        assert self.f is not None, "FunctioProvider created with None!"
        self._dependencies = extract_dependency(self.f)

    @property
    def provider(self):
        return self.f


# and anything that converts to provider


def ensure_provider_name(f):
    if not f.__name__.startswith("provide_"):
        f.__name__ = "provide_" + f.__name__
    return f


def ensure_self_arg(func, fname: str = None):
    """
    adds a self arg in to func signature if not present
    :param func:
    :return:
    """
    assert func is not None
    argspec = inspect.getfullargspec(func)
    if argspec.args and argspec.args[0] == "self":
        return func
    assert "self" not in argspec.args, f"{argspec.args}"
    # source = inspect.getsource(func)
    module = inspect.getmodule(func)
    # module can be None if func is from the pickled.
    if fname is None:
        try:
            lines, ln = inspect.getsourcelines(func)
            fn = func.__name__.replace("<", "").replace(">", "")
            fname = f"""{module.__name__.replace(".", "_")}_ln_{ln}_{fn}"""
        except OSError as ose:
            fn = func.__name__.replace("<", "").replace(">", "")
            fname = f"""{module.__name__.replace(".", "_")}_{fn}"""
        except AttributeError as ae:
            logger.warning(f"could not set reasonable func name. better provide fname as argument!")
            fname = f"""unknwon_module_function"""

    assert not isinstance(argspec.args, str)
    signature = f"""{fname}({" ,".join(["self"] + (argspec.args or []))})"""

    # logger.error(source)

    def impl(self, *args, **kwargs):
        return func(*args, **kwargs)  # gets multiple values for self

    assert len(argspec.kwonlyargs) == 0, "providing method cannot have any kwonly args"
    return create_function(signature, impl)


@dataclass
class PinjectBind(Bind):
    kwargs: Dict

    def __hash__(self):
        return hash(frozendict(**self.kwargs))

    def to_pinject_binding(self) -> Union[PinjectConfigure, PinjectProvider]:
        return PinjectConfigure(self.kwargs)

@dataclass
class MetaBind(Bind):
    src:Bind
    metadata:dict
    def to_pinject_binding(self) -> Union[PinjectConfigure, PinjectProvider]:
        return self.src.to_pinject_binding()



import abc
import functools
import inspect
import sys
from dataclasses import dataclass
from pprint import pformat
from typing import List, Generic, Mapping, Union, Callable, TypeVar, Tuple, Set, Any, Dict

import makefun
from loguru import logger
from makefun import create_function

T, U = TypeVar("T"), TypeVar("U")

A = TypeVar("A")
B = TypeVar("B")


@dataclass
class FrameInfo:
    filename: str
    line_number: int
    function_name: str
    sources: List[str]
    line_idx_in_sources: int


def get_frame_info(stack_idx) -> Union[None, FrameInfo]:
    try:
        return FrameInfo(*inspect.getframeinfo(inspect.stack(0)[stack_idx][0]))
    except Exception as e:
        return None


def partialclass(name, cls, *args, **kwds):
    """
    copied from https://stackoverflow.com/questions/38911146/python-equivalent-of-functools-partial-for-a-class-constructor
    :param name:
    :param cls:
    :param args:
    :param kwds:
    :return:
    """
    new_cls = type(name, (cls,), {
        '__init__': functools.partialmethod(cls.__init__, *args, **kwds)
    })

    # The following is copied nearly ad verbatim from `namedtuple's` source.
    """
    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in enviroments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython).
    """
    try:
        new_cls.__module__ = sys._getframe(1).f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass

    return new_cls


class Injected(Generic[T], metaclass=abc.ABCMeta):
    """
    this class is actually an abstraction of fucntion partial application.
    """

    @staticmethod
    def partial(target_function: Callable, **injection_targets: "Injected") -> "Injected[Callable]":
        """
        use this to partially inject specified params, and leave the other parameters to be provided after injection is resolved
        :param target_function: Callable
        :param injection_targets: specific parameters to make injected automatically
        :return: Injected[Callable[(params which were not specified in injection_targets)=>Any]]
        """
        # how can I partially apply class constructor?
        argspec = inspect.getfullargspec(target_function)
        remaining_arg_names = argspec.args
        if "self" in remaining_arg_names:
            remaining_arg_names.remove("self")
        #logger.info(f"partially applying {injection_targets}")
        #logger.info(f"original args:{remaining_arg_names}")

        for injected in injection_targets.keys():
            remaining_arg_names.remove(injected)

        def makefun_impl(kwargs):
            #logger.info(f"partial injection :{pformat(kwargs)}")

            def inner(*_args):
                # user calls with both args or kwargs so we need to handle both.
                assert len(_args) == len(remaining_arg_names),\
                    f"partially applied injected function is missing some of positional args! {remaining_arg_names} for {_args}"
                call_kwargs = dict(zip(remaining_arg_names, _args))
                #logger.info(f"partial injection call :{pformat(call_kwargs)}")
                full_kwargs = {**kwargs, **call_kwargs}
                return target_function(**full_kwargs)

            return inner
        injected_kwargs = Injected.dict(**injection_targets)
        injected_factory = Injected.bind(makefun_impl,kwargs=injected_kwargs)
        return injected_factory


    @staticmethod
    def inject_except(target_function, *whitelist: str) -> "Injected[Callable]":
        """
        :param target_function:
        :param whitelist: name of arguments which should not be injected by DI.
        :return: Injected[Callable[(whitelisted args)=>Any]]
        """
        argspec = inspect.getfullargspec(target_function)
        args_to_be_injected = [a for a in argspec.args if a not in whitelist and a != "self"]
        return Injected.partial(target_function, *args_to_be_injected)

    @staticmethod
    def bind(_target_function_, **kwargs_mapping: Union[str, type, Callable, "Injected"]) -> "Injected":
        if isinstance(_target_function_,Injected):
            _target_function_ = _target_function_.get_provider()
        return InjectedFunction(
            target_function=_target_function_,
            kwargs_mapping=kwargs_mapping
        )

    @staticmethod
    def direct(_target_function, **kwargs) -> "Injected":
        """
        uses Injected.pure by default unless the parameter is an instance of Injected.
        :param _target_function:
        :param kwargs:
        :return:
        """
        en_injected = {k: Injected.pure(v) for k, v in kwargs.items() if not isinstance(v, Injected)}
        already_injected = {k: v for k, v in kwargs.items() if isinstance(v, Injected)}
        return Injected.bind(_target_function, **en_injected, **already_injected)

    def _faster_get_fname(self):
        try:
            frame = sys._getframe().f_back.f_back.f_back.f_back
            mod = frame.f_globals["__name__"]
            name = frame.f_lineno
            return f"{mod.replace('.', '_')}_L_{name}".replace("<", "__").replace(">", "__")
        except Exception as e:
            from loguru import logger
            logger.warning(f"failed to get name of the injected location.")
            return f"__unknown_module__maybe_due_to_pickling__"

    def __init__(self, init_stack):
        self.fname = self._faster_get_fname()

    @abc.abstractmethod
    def dependencies(self) -> Set[str]:
        pass

    def get_signature(self):
        sig = f"""{self.fname}({",".join(self.dependencies())})"""
        # logger.warning(sig)
        return sig

    @abc.abstractmethod
    def get_provider(self):
        """
        :return: a provider
        """

    def map(self, f: Callable[[T], U]) -> 'Injected[U]':
        return MappedInjected(4, self, f)

    @staticmethod
    def from_impl(impl: Callable, dependencies: Set[str]):
        return GeneratedInjected(impl, dependencies, init_stack=4)

    @staticmethod
    def pure(value):
        return InjectedPure(value, init_stack=4)

    @staticmethod
    def by_name(name: str):
        return InjectedByName(name, init_stack=4)

    def zip(self, other: "Injected[U]") -> "Injected[Tuple[T,U]]":
        return ZippedInjected(self, other, init_stack=4)

    @staticmethod
    def mzip(*srcs: "Injected"):
        return MZippedInjected(4, *srcs)

    # this is ap of applicative functor.
    def apply_injected_function(self, other: "Injected[Callable[[T],U]]") -> "Injected[U]":
        return self.zip(other).map(
            lambda t: t[1](t[0])
        )

    @staticmethod
    def dict(**kwargs: "Injected") -> "Injected[Dict]":
        keys = list(kwargs.keys())
        return Injected.mzip(*[kwargs[k] for k in keys]).map(lambda t: {k: v for k, v in zip(keys, t)})


class GeneratedInjected(Injected):
    """creates Injected from dependencies and funct(**kwargs) signature"""

    def __init__(self, impl: Callable, dependencies: Set[str], init_stack):
        super().__init__(init_stack)
        self.impl = impl
        self._dependencies = dependencies

    def dependencies(self) -> Set[str]:
        return self._dependencies

    def get_provider(self):
        return create_function(self.get_signature(), func_impl=self.impl)


class MappedInjected(Injected):
    src: Injected[T]
    f: Callable[[T], U]

    def __init__(self, stack, src, f):
        super(MappedInjected, self).__init__(stack)
        self.src = src
        self.f = f

    def dependencies(self) -> Set[str]:
        return self.src.dependencies()

    def get_provider(self):
        def impl(**kwargs):
            assert self.dependencies() == set(
                kwargs.keys())  # this is fine but get_provider returns wrong signatured func
            tmp = self.src.get_provider()(**kwargs)
            return self.f(tmp)

        return create_function(self.get_signature(), func_impl=impl)


class MapWithExtras(Injected):
    src: Injected[T]
    f: Callable[[T, Any], U]
    extras: Set[str]

    def __init__(self, stack, src, f, extras: Set[str]):
        super(MappedInjected, self).__init__(stack)
        self.src = src
        self.f = f
        self.extras = extras

    def dependencies(self) -> Set[str]:
        return self.src.dependencies() | self.extras

    def get_provider(self):
        def impl(**kwargs):
            src_deps = {k: kwargs[k] for k in self.src.dependencies()}
            extras = {k: kwargs[k] for k in self.extras}
            tmp = self.src.get_provider()(**src_deps)
            return self.f(tmp, **extras)

        return create_function(self.get_signature(), func_impl=impl)


def extract_dependency_including_self(f: Union[type, Callable]):
    if isinstance(f, type):  # it's a constructor and we must use __init__ for backward compatibility.
        argspec = inspect.getfullargspec(f.__init__)
        return set(argspec.args)
    elif isinstance(f, Callable):
        argspec = inspect.getfullargspec(f)
        return set(argspec.args)
    else:
        raise RuntimeError(f"input:{f} is not either type or Callable")


def extract_dependency(dep: Union[str, type, Callable, Injected]) -> Set[str]:
    if isinstance(dep, str):
        return {dep}
    elif isinstance(dep, type):  # it's a constructor and we must use __init__ for backward compatibility.
        argspec = inspect.getfullargspec(dep.__init__)
        return set(argspec.args) - {'self'}
    elif isinstance(dep, Callable):
        try:
            argspec = inspect.getfullargspec(dep)
        except Exception as e:
            raise e

        return set(argspec.args) - {'self'}
    elif isinstance(dep, Injected):
        return dep.dependencies()
    else:
        raise RuntimeError(f"dep must be either str/type/Callable/Injected. got {type(dep)}")


def solve_injection(dep: Union[str, type, Callable, Injected], kwargs: dict):
    if isinstance(dep, str):
        return kwargs[dep]
    elif isinstance(dep, (type, Callable)):
        return dep(**{k: kwargs[k] for k in extract_dependency(dep)})
    elif isinstance(dep, Injected):
        return solve_injection(dep.get_provider(), kwargs)
    else:
        raise RuntimeError(f"dep must be one of str/type/Callable/Injected. got {type(dep)}")


def combine_image_store(a, b):
    # do anything
    return a + b


def assert_kwargs_type(v):
    if isinstance(v, str):
        return
    if isinstance(v, type):
        return
    if isinstance(v, Callable):
        return
    if isinstance(v, Injected):
        return
    else:
        raise TypeError(f"{type(v)} is not any of [str,type,Callable,Injected]")


class InjectedFunction(Injected[T]):
    # since the behavior differs in classes extending Generic[T]
    def __init__(self,
                 target_function: Callable,
                 kwargs_mapping: Mapping[str, Union[str, type, Callable, Injected]]
                 ):
        super().__init__(init_stack=4)
        assert callable(target_function)
        self.target_function = target_function
        self.kwargs_mapping = kwargs_mapping
        for k, v in self.kwargs_mapping.items():
            assert v is not None, f"injected got None binding for key:{k}"
            assert_kwargs_type(v)

        org_deps = extract_dependency(self.target_function)
        # logger.info(f"tgt:{target_function} original dependency:{org_deps}")
        missings = {d for d in org_deps if d not in self.kwargs_mapping}
        # logger.info(f"now missing {missings}")
        # logger.warning(f"created InjectedFunction:{inspect.signature(target_function)}")
        # assert "self" not in inspect.signature(target_function).parameters
        self.missings = missings

    def get_provider(self):
        signature = self.get_signature()

        def impl(**kwargs):
            deps = dict()
            for mdep in self.missings:
                deps[mdep] = solve_injection(mdep, kwargs)
            for k, dep in self.kwargs_mapping.items():
                deps[k] = solve_injection(dep, kwargs)
            # logger.info(f"calling function:{self.target_function.__name__}{inspect.signature(self.target_function)}")
            return self.target_function(**deps)

        # you have to add a prefix 'provider'""
        return create_function(func_signature=signature, func_impl=impl)

    def dependencies(self) -> Set[str]:
        # ahhhh this recursively demands for injection.
        # we need to distinguish what and what not to recursively inject
        res = set()
        for mdep in self.missings:
            d = extract_dependency(mdep)
            res |= d
            # logger.info(f"deps of missing:{d}")
        for k, dep in self.kwargs_mapping.items():
            d = extract_dependency(dep)
            res |= d
            # logger.info(f"deps of dependency({k}):{d}")
        return res

    #def __str__(self):
    #    return f"""InjectedFunction(target={self.target_function},kwargs_mapping={self.kwargs_mapping})"""


class InjectedPure(Injected[T]):
    def __init__(self, value, init_stack):
        super().__init__(init_stack)
        self.value = value

    def dependencies(self) -> Set[str]:
        return set()

    def get_provider(self):
        return create_function(func_signature=self.get_signature(), func_impl=lambda: self.value)


class InjectedByName(Injected[T]):
    def __init__(self, name, init_stack):
        super().__init__(init_stack)
        self.name = name

    def dependencies(self) -> Set[str]:
        return {self.name}

    def get_provider(self):
        return create_function(func_signature=self.get_signature(), func_impl=lambda **kwargs: kwargs[self.name])


class ZippedInjected(Injected[Tuple[A, B]]):
    def __init__(self, a: Injected[A], b: Injected[B], init_stack):
        super().__init__(init_stack)
        self.a = a
        self.b = b

    def dependencies(self) -> Set[str]:
        return set(self.a.dependencies() | self.b.dependencies())

    def get_provider(self):
        def impl(**kwargs):  # can we pickle this though?
            # logger.info(f"providing from ZippedInjected!:{kwargs}")
            # logger.info(f"a:{self.a}")
            # logger.info(f"b:{self.b}")
            a_kwargs = {k: kwargs[k] for k in self.a.dependencies()}
            b_kwargs = {k: kwargs[k] for k in self.b.dependencies()}
            # embed()
            a = self.a.get_provider()(**a_kwargs)
            # logger.info(f"a:{a}")
            b = self.b.get_provider()(**b_kwargs)
            # logger.info(f"b:{b}")
            # logger.info((a, b))
            return a, b

        signature = self.get_signature()
        # logger.info(f"created signature:{signature} for ZippedInjected")
        return create_function(func_signature=signature, func_impl=impl)



class MZippedInjected(Injected):
    def __init__(self, init_stack, *srcs: Injected):
        super().__init__(init_stack)
        self.srcs = srcs

    def dependencies(self) -> Set[str]:
        res = set()
        for s in self.srcs:
            res |= s.dependencies()
        return res

    def get_provider(self):
        def impl(**kwargs):  # can we pickle this though?
            each_kwargs = [{k: kwargs[k] for k in s.dependencies()} for s in self.srcs]
            res = []
            for s in self.srcs:
                r = s.get_provider()(**{k: kwargs[k] for k in s.dependencies()})
                res.append(r)
            return tuple(res)

        signature = self.get_signature()
        from loguru import logger
        logger.info(f"created signature:{signature} for MZippedInjected")
        return create_function(func_signature=signature, func_impl=impl)

def _injected_factory(**targets: Injected):
    def _impl(f):
        return Injected.partial(f, **targets)

    return _impl


def injected_factory(f):
    """
    any args starting with "_" is considered to be injected.
    :param f:
    :return:
    """
    sig: inspect.Signature = inspect.signature(f)
    tgts = dict()
    for k in sig.parameters.keys():
        if k.startswith("_"):
            tgts[k] = Injected.by_name(k[1:])
    return _injected_factory(**tgts)(f)


from functools import partial
from typing import Callable, Any

from loguru import logger

from pinject_design.di.injected import Injected
from pinject_design.di.util import Design


def _factory(a, b, x):
    return a + b + x


def test_partial():
    d = Design().bind_instance(
        a=1,
        b=2
    ).bind_provider(
        f=Injected.partial(_factory, "a", "b")
    )
    f: Callable[[int], Any] = d.provide("f")
    applied: Callable[[int], Any] = partial(_factory, 0, 0)

    assert partial(_factory, 0, b=0)(x=1) == 1
    fx = f(3)
    logger.info(f)
    assert fx == 6



import inspect
from types import FunctionType
from typing import Union, Type, Callable, TypeVar, List

from loguru import logger
from makefun import create_function
from pinject.bindings import default_get_arg_names_from_class_name
from pinject.errors import NothingInjectableForArgError
from pinject.object_graph import ObjectGraph
from pinject_design.di.injected import Injected
from pinject_design.exceptions import DependencyResolutionFailure

T = TypeVar("T")


class MissingDependencyException(Exception):
    @staticmethod
    def create_message(deps: List[DependencyResolutionFailure]):
        msgs = [item.explanation_str() for item in deps]
        lines = '\n'.join(msgs)
        return f"Missing dependency. failures:\n {lines}."

    @staticmethod
    def create(deps: List[DependencyResolutionFailure]):
        return MissingDependencyException(MissingDependencyException.create_message(deps))


class ExtendedObjectGraph:
    """
    an object graph which can also provide instances based on its name not only from class object.
    """

    def __init__(self, design: "Design", src: ObjectGraph):
        self.design = design
        self.src = src

    def _provide(self, target: Union[str, Type[T], Injected[T], Callable]) -> Union[object, T]:
        if isinstance(target, str):
            assert target != "self", f"provide target:{target}"
            code = compile(f"""def __init__(self,{target}):self.res={target}""", "<string>", "exec")
            fn = FunctionType(code.co_consts[0], globals(), "__init__")
            Request = type("Request", (object,), dict(__init__=fn))
            return self.src.provide(Request).res
        elif isinstance(target, type):
            return self.src.provide(target)
        elif isinstance(target, Injected):
            deps = target.dependencies()
            if 'self' in deps:
                deps.remove('self')
            signature = f"""__init__(self,{','.join(deps)})"""

            def impl(self, **kwargs):
                self.data = target.get_provider()(**kwargs)

            __init__ = create_function(signature, func_impl=impl)
            Request = type("Request", (object,), dict(__init__=__init__))
            return self.src.provide(Request).data
        elif isinstance(target, Callable):
            return self.run(target)
        else:
            raise TypeError(f"target must be either class or a string or Injected. got {target}")

    def provide(self, target: Union[str, Type[T], Injected[T], Callable]) -> Union[object, T]:
        try:
            return self._provide(target)
        except NothingInjectableForArgError as e:
            # preventing circular import
            from pinject_design.visualize_di import DIGraph
            match target:
                case type():
                    deps = [default_get_arg_names_from_class_name(target.__name__)[0]]
                case Injected():
                    deps = target.dependencies()
                case str():
                    deps = [target]
                case other:
                    raise e

            missings = DIGraph(self.design).find_missing_dependencies(deps)
            if missings:
                for missing in missings:
                    logger.error(f"failed to find dependency:{missing}")
                raise MissingDependencyException.create(missings)
            else:
                raise e

    def run(self, f):
        argspec = inspect.getfullargspec(f)
        assert "self" not in argspec.args, f"self in {argspec.args}, of {f}"
        # logger.info(self)
        assert argspec.varargs is None
        kwargs = {k: self.provide(k) for k in argspec.args}
        return f(**kwargs)

    def __repr__(self):
        return f"ExtendedObjectGraph of a design:\n{self.design}"

    def __getitem__(self, item):
        return self.provide(item)



from dataclasses import dataclass
from typing import List, Any


@dataclass
class DependencyResolutionFailure:
    key: str
    trace: List[str]
    cause: Any

    def trace_str(self):
        return ' => '.join(self.trace)

    def explanation_str(self):
        return f"failed to find dependency: {self.key} at {self.trace_str()}"

    def __repr__(self):
        return f"DependencyResolutionFailure(key:{self.key}\t,trace:{self.trace_str()},cause: ({self.cause})"


class _MissingDepsError(RuntimeError):
    def __init__(self, msg: str, name: str, trace: List[str]):
        super().__init__(msg)
        self.name = name
        self.trace = trace.copy()

    def __getstate__(self):
        return self.msg, self.name, self.trace

    def __setstate__(self, data):
        self.msg, self.name, self.trace = data


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pinject_design', 'pinject_design.di', 'pinject_design.viz']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle',
 'cytoolz',
 'expression',
 'loguru',
 'makefun',
 'pampy',
 'pinject',
 'returns',
 'tabulate']

setup_kwargs = {
    'name': 'pinject-design',
    'version': '0.1.30',
    'description': 'immutable wrapper for pinject',
    'long_description': None,
    'author': 'proboscis',
    'author_email': 'nameissoap@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
